from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ChefScript")
except PackageNotFoundError:
    __version__ = "0.0.0"

MAX_TERMINTAL_WIDTH = 80


def pretty_str(s: str) -> str:
    return s.center(MAX_TERMINTAL_WIDTH // 2).center(MAX_TERMINTAL_WIDTH, "-")


def index_to_position(s: str, index: int) -> Position:
    """
    Returns (line_number, col) of `index` in `s`.

    https://stackoverflow.com/a/66443805/11946514
    """
    if not len(s):
        return Position(-1, -1)
    sp = s[: index + 1].splitlines(keepends=True)
    return Position(len(sp), len(sp[-1]))


class Position:
    def __init__(self, line: int, col: int):
        self.line = line
        self.col = col

    def __str__(self):
        return f"{self.line}:{self.col}"

    def __repr__(self):
        return f"{self.__class__.__name__}(line={self.line}, col={self.col})"

    def __bool__(self):
        return self.line > 0 and self.col > 0


class ChefScriptException(BaseException):
    ...


class ChefScriptInternalError(ChefScriptException):
    def __init__(self, msg: str, fn: str):
        self.msg = msg
        self.fn = fn

    def __str__(self):
        return f"In {self.fn}: internal error: {self.msg}"


class ChefScriptErrorWithPosition(ChefScriptException):
    def __init__(self, msg: str, fn: str, pos: Position):
        self.msg = msg
        self.fn = fn
        self.pos = pos


class ChefScriptKeyboardInterrupt(ChefScriptErrorWithPosition):
    def __str__(self):
        if self.msg:
            return f"In {self.fn}:{self.pos}: keyboard interrupt: {self.msg}"
        elif self.pos:
            return f"In {self.fn}:{self.pos}: keyboard interrupt"
        else:
            return f"In {self.fn}: keyboard interrupt"


class ChefScriptSyntaxError(ChefScriptErrorWithPosition):
    def __str__(self):
        return f"In {self.fn}:{self.pos}: syntax error: {self.msg}"


class ChefScriptRuntimeError(ChefScriptErrorWithPosition):
    def __str__(self):
        return f"In {self.fn}:{self.pos}: runtime error: {self.msg}"



from __future__ import annotations

from copy import copy
from typing import overload

from quantities import Quantity as PQuantity

from .utils import Real

__all__ = ["Quantity", "Ingredient"]


class Quantity:
    """
    A simple quantity class to hold a value and a unit.

    It is intended to hold the weights and volumes of ingredients in cooking,
    so it doesn't support complex units (multiplication by another quantity),
    but it does support multiplication by bare numbers.
    """

    quantity: PQuantity

    def __init__(self, value: Real, unit: str) -> None:
        self.quantity = PQuantity(value, self._normalize_unit(unit))

    def __copy__(self) -> Quantity:
        return Quantity(self.value, self.unit)

    def __repr__(self) -> str:
        return f"{self.value:.3f} {self.unit}"

    def __add__(self, other) -> Quantity:
        if isinstance(other, Quantity):
            other_copy = copy(other)
            other_copy.rescale(self.unit)
            return Quantity(self.value + other_copy.value, self.unit)
        else:
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
            )

    def __sub__(self, other) -> Quantity:
        if isinstance(other, Quantity):
            other_copy = copy(other)
            other_copy.rescale(self.unit)
            return Quantity(self.value - other_copy.value, self.unit)
        else:
            raise TypeError(
                f"unsupported operand type(s) for -: '{type(self)}' and '{type(other)}'"
            )

    def __mul__(self, other) -> Quantity:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return Quantity(self.value * other, self.unit)
        else:
            raise TypeError(
                f"unsupported operand type(s) for *: '{type(self)}' and '{type(other)}'"
            )

    def __rmul__(self, other) -> Quantity:
        return self.__mul__(other)

    @overload
    def __truediv__(self, other: float) -> Quantity:
        ...

    @overload
    def __truediv__(self, other: Quantity) -> float:
        ...

    def __truediv__(self, other) -> Quantity | float:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return self.__mul__(1 / other)
        elif isinstance(other, Quantity):
            other_copy = copy(other)
            other_copy.rescale(self.unit)
            return self.value / other_copy.value
        else:
            raise TypeError(
                f"unsupported operand type(s) for /: '{type(self)}' and '{type(other)}'"
            )

    @staticmethod
    def _normalize_unit(unit: str) -> str:
        return unit.replace(".", "").replace(" ", "_").replace("^", "**")

    def rescale(self, unit: str) -> None:
        self.quantity = self.quantity.rescale(self._normalize_unit(unit))

    @property
    def value(self) -> float:
        return float(self.quantity.magnitude)

    @property
    def unit(self) -> str:
        return str(self.quantity.dimensionality).replace("_", " ")


class Ingredient:
    def __init__(self, name: str, quantity: Quantity) -> None:
        self.name = name
        self.quantity = quantity

    def __repr__(self) -> str:
        return f"{self.quantity} of {self.name}"

    def __add__(self, other) -> Ingredient:
        if isinstance(other, Ingredient):
            if self.name != other.name:
                raise ValueError("Ingredients must have the same name")
            return Ingredient(self.name, self.quantity + other.quantity)
        else:
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
            )

    def __sub__(self, other) -> Ingredient:
        if isinstance(other, Ingredient):
            if self.name != other.name:
                raise ValueError("Ingredients must have the same name")
            return Ingredient(self.name, self.quantity - other.quantity)
        else:
            raise TypeError(
                f"unsupported operand type(s) for -: '{type(self)}' and '{type(other)}'"
            )

    def __mul__(self, other) -> Ingredient:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return Ingredient(self.name, self.quantity * other)
        else:
            raise TypeError(
                f"unsupported operand type(s) for *: '{type(self)}' and '{type(other)}'"
            )

    def __rmul__(self, other) -> Ingredient:
        return self.__mul__(other)

    def __truediv__(self, other) -> Ingredient:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return self.__mul__(1 / other)
        else:
            raise TypeError(
                f"unsupported operand type(s) for /: '{type(self)}' and '{type(other)}'"
            )



from argparse import ArgumentParser
from pathlib import Path
from sys import stderr

from .interpreter import ChefScriptInterpreter
from .utils import ChefScriptKeyboardInterrupt, __version__, pretty_str


def main():
    parser = ArgumentParser(
        description="Run a ChefScript programme. "
        "When no file is specified, ChefScript will run in interactive mode."
    )
    parser.add_argument(
        "filename",
        type=Path,
        help="Enter the ChefScript file to run",
        nargs="?",
        default=None,
    )
    args = parser.parse_args()

    interpreter = ChefScriptInterpreter()

    try:
        if args.filename is None or args.filename == "-":
            print(pretty_str(f"This is ChefScript {__version__}"))
            interpreter.interpret_stdin()
        else:
            try:
                interpreter.interpret_file(args.filename.resolve())
            except FileNotFoundError:
                print(
                    f"No such file or directory: '{args.filename}'",
                    file=stderr,
                )

    except KeyboardInterrupt:
        keyboard_interrupt = ChefScriptKeyboardInterrupt(
            "",
            interpreter.filename,
            interpreter.pos,
        )
        print(keyboard_interrupt, file=stderr)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections import OrderedDict, deque
from typing import Optional, Sequence

from .ingredient import Ingredient, Quantity
from .utils import TAB, Real

__all__ = ["Recipe"]


class Recipe:
    def __init__(
        self,
        name: str,
        instructions: Sequence[tuple[Ingredient | Recipe, Optional[str]]],
    ) -> None:
        self.name = name
        self.instructions = instructions

    def __repr__(self) -> str:
        str_builder = [f"{self.name}"]
        for instruction in self.instructions:
            ingredient_or_recipe = instruction[0]
            if isinstance(ingredient_or_recipe, Ingredient):
                item = str(ingredient_or_recipe)
            else:
                item = ingredient_or_recipe.name
            str_builder.append(
                f"{TAB}{item}{'' if instruction[1] is None else f' ({instruction[1]})'}"
            )
        return "\n".join(str_builder)

    def __mul__(self, other) -> Recipe:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return Recipe(self.name, [(i[0] * other, i[1]) for i in self.instructions])
        else:
            raise TypeError(
                f"unsupported operand type(s) for *: '{type(self)}' and '{type(other)}'"
            )

    def __rmul__(self, other) -> Recipe:
        return self.__mul__(other)

    def __truediv__(self, other) -> Recipe:
        if isinstance(other, Real):  # type: ignore[misc, arg-type]
            return self.__mul__(1 / other)
        else:
            raise TypeError(
                f"unsupported operand type(s) for /: '{type(self)}' and '{type(other)}'"
            )

    @property
    def ingredients(self) -> list[Ingredient]:
        ingredients: list[tuple[str, Quantity]] = []
        referenced_recipe_names: set[str] = set()
        for instruction in self.instructions:
            ingredient_or_recipe = instruction[0]
            if isinstance(ingredient_or_recipe, Ingredient):
                ingredient = ingredient_or_recipe
                ingredients.append((ingredient.name, ingredient.quantity))
            else:
                recipe = ingredient_or_recipe
                if recipe.name not in referenced_recipe_names:
                    referenced_recipe_names.add(recipe.name)
                    for ingredient in recipe.ingredients:
                        ingredients.append((ingredient.name, ingredient.quantity))
        ingredients_dict: dict[str, Quantity] = {}
        for name, quantity in ingredients:
            if name in ingredients_dict:
                ingredients_dict[name] += quantity
            else:
                ingredients_dict[name] = quantity
        return sorted(
            (Ingredient(name, quantity) for name, quantity in ingredients_dict.items()),
            key=lambda i: i.name,
        )

    @property
    def pretty_str(self) -> str:
        str_builder = deque([f"Recipe for {self.name}:"])
        referenced_recipes: OrderedDict[str, Recipe] = OrderedDict()

        str_builder.append(f"{TAB}Summary of ingredients:")
        for ingredient in self.ingredients:
            str_builder.append(f"{TAB*2}{ingredient}")

        str_builder.append(f"{TAB}Instructions:")
        for instruction in self.instructions:
            ingredient_or_recipe = instruction[0]
            if isinstance(ingredient_or_recipe, Ingredient):
                item = str(ingredient_or_recipe)
            else:
                referenced_recipes.update(
                    {ingredient_or_recipe.name: ingredient_or_recipe}
                )
                item = ingredient_or_recipe.name
            str_builder.append(
                f"{TAB*2}{item}"
                f"{'' if instruction[1] is None else f' ({instruction[1]})'}"
            )

        for name, recipe in referenced_recipes.items():
            str_builder.appendleft(recipe.pretty_str)

        return "\n".join(str_builder)



from __future__ import annotations

from typing import NamedTuple

from pyparsing import (
    Group,
    IndentedBlock,
    Keyword,
    OneOrMore,
    Optional,
    ParserElement,
    Regex,
    Suppress,
    ZeroOrMore,
    pyparsing_common,
)
from regex import compile as regex_compile

from pychef import (
    Ingredient as PychefIngredient,
    Quantity as PychefQuantity,
    Recipe as PychefRecipe,
)

ParserElement.set_default_whitespace_chars(" \t")


class Comment(str):
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"


class Cook(NamedTuple):
    recipe_name: str
    scale: float | PychefIngredient | None
    idx: int  # traceback information


class Parser:
    parser: ParserElement

    @classmethod
    def parse(cls, string):
        return cls.parser.parse_string(string, parse_all=True)


class Keywords(Parser):
    NEWLINE = Suppress("\n")
    COMMENT_START = Suppress("(")
    COMMENT_END = Suppress(")")
    OF = Suppress(Keyword("of"))
    COOK = Suppress(Keyword("cook"))
    FOR = Suppress(Keyword("for"))
    TIMES = Suppress(Keyword("times"))
    WITH = Suppress(Keyword("with"))

    parser = NEWLINE | COMMENT_START | COMMENT_END | OF | COOK | FOR | TIMES | WITH


class NumberParser(Parser):
    r"""
    ``<number> ::= <single_number> | <single_number> <number>``

    ``<single_number> ::= <fraction> | <signed_float>``

    ``<fraction> ::= <signed_float>"/"<signed_float>`` (no spaces)

    ``<signed_float>`` is defined by the Regex ``[+-]?\d+\.?\d*([eE][+-]?\d+)?``
    """

    signed_float = pyparsing_common.fnumber.copy()

    fraction = signed_float + "/" + signed_float
    fraction.add_condition(
        lambda t: float(t[2]) != 0, message="Denominator must not be zero", fatal=True
    )
    fraction.add_parse_action(lambda loc, t: float(t[0]) / float(t[2]))

    single_number = fraction | signed_float
    single_number.leave_whitespace(recursive=False)

    number = OneOrMore(single_number)
    number.ignore_whitespace(recursive=False)
    number.add_parse_action(lambda loc, t: sum(t))

    parser = number
    """Returns the float"""


class VariableNameParser(Parser):
    r"""
    ``<variable_name> ::= <word> | <word> <variable_name>``

    ``<word>`` is a string any character except control characters and whitespace,
    defined by the Regex ``[^\p{C}\p{Z}]+``
    """

    word = Regex(regex_compile(r"[^\p{C}\p{Z}]+"))

    variable_name = OneOrMore(word, stop_on=Keywords.parser)
    variable_name.add_parse_action(lambda loc, t: " ".join(t))

    parser = variable_name
    """Returns the variable, normalised by removing duplicate whitespaces"""


class QuantityParser(Parser):
    """
    ``<quantity> ::= <number> <unit>`` (optional space)

    ``<unit> ::= <variable_name>``
    """

    @staticmethod
    def quantity_condition(t):
        try:
            PychefQuantity(t[0], t[1])
        except Exception:
            return False
        return True

    @staticmethod
    def quantity_parse_action(loc, t):
        return PychefQuantity(t[0], t[1])

    unit = VariableNameParser.parser.copy()
    unit.add_parse_action(lambda loc, t: t[0])

    quantity = NumberParser.parser + unit
    quantity.add_condition(quantity_condition, message="Invalid quantity", fatal=True)
    quantity.add_parse_action(quantity_parse_action)

    parser = quantity
    """Returns the quantity"""


class CommentParser(Parser):
    """
    ``<comment> ::= <comment_start> <comment_text> <comment_end>``

    ``<comment_start> ::= "("``

    ``<comment_end> ::= ")"``

    ``<comment_text>`` is string of any character except ``)``,
    defined by the Regex ``[^)]+``
    """

    comment_text = Regex(regex_compile(r"[^)]+"))

    comment = Keywords.COMMENT_START + comment_text + Keywords.COMMENT_END

    suppressed_comment = Suppress(comment)

    parser = comment
    """Returns the string of the comment text"""


class IngredientParser(Parser):
    """
    ``<ingredient> ::= <quantity> "of" <variable_name>``
    """

    ingredient = (
        QuantityParser.parser + Suppress(Keyword("of")) + VariableNameParser.parser
    )
    ingredient.add_parse_action(lambda loc, t: PychefIngredient(t[1], t[0]))

    parser = ingredient
    """Returns the ingredient"""


class RecipeParser(Parser):
    """
    ``<recipe> ::= <recipe_name> NEWLINE INDENT <recipe_body>``

    ``<recipe_name> ::= <variable_name>``

    ``<recipe_body> ::= <recipe_line> NEWLINE | <recipe_line> NEWLINE <recipe_body>``

    ``<recipe_line> ::= <comment> | <ingredient> <comment> | ``

    (``NEWLINE`` is allowed to be repeated)
    """

    @staticmethod
    def recipe_parse_action(loc, t):
        name = t[0]
        instructions = t[1]

        for i, instruction in enumerate(instructions):
            if isinstance(instruction[0], str):
                instructions[i][0] = PychefRecipe(instructions[i][0], [])
            if len(instruction) != 2:
                instruction.append(None)

        recipe = PychefRecipe(name, instructions)
        recipe.idx = loc  # type: ignore
        return recipe

    recipe_line = CommentParser.suppressed_comment | Group(
        Optional(IngredientParser.parser | VariableNameParser.parser)
        + Optional(CommentParser.parser)
    )

    recipe_body = IndentedBlock(
        ZeroOrMore(Keywords.NEWLINE) + recipe_line + OneOrMore(Keywords.NEWLINE)
    )

    recipe = (
        VariableNameParser.parser
        + Optional(CommentParser.suppressed_comment)
        + OneOrMore(Keywords.NEWLINE)
        + recipe_body
    )
    recipe.add_parse_action(recipe_parse_action)

    parser = recipe


class CookStatementParser(Parser):
    """
    ``<cook_statement> ::= "cook" <recipe_name>
                         | "cook" <recipe_name> <times>
                         | "cook" <recipe_name> <with>``

    ``<times> ::= "for" <number> "times"``

    ``<with> ::= "with" <ingredient>``
    """

    times = Keywords.FOR + NumberParser.parser + Keywords.TIMES

    with_ = Keywords.WITH + IngredientParser.parser

    cook_statement = Keywords.COOK + VariableNameParser.parser + Optional(times | with_)
    cook_statement.add_parse_action(
        lambda loc, t: Cook(t[0], t[1], loc) if len(t) == 2 else Cook(t[0], None, loc)
    )

    parser = cook_statement


class ChefScriptParser(Parser):
    """
    ``<chef_script> ::= <stmt> | <stmt> NEWLINE <chef_script>``

    ``<stmt> ::= <recipe> | <cook_statement>``

    (NEWLINE is allowed to be repeated)
    """

    stmt = CommentParser.suppressed_comment | Group(
        (RecipeParser.parser | CookStatementParser.parser)
        + Optional(CommentParser.suppressed_comment)
    )

    chef_script = (
        ZeroOrMore(Keywords.NEWLINE)
        + OneOrMore(stmt + ZeroOrMore(Keywords.NEWLINE))
        + ZeroOrMore(Keywords.NEWLINE)
    )

    parser = chef_script



from collections import OrderedDict
from pathlib import Path
from sys import stderr, stdin

from pyparsing.exceptions import ParseBaseException

from pychef import Recipe as PychefRecipe

from .parser import ChefScriptParser, Cook
from .utils import (
    MAX_TERMINTAL_WIDTH,
    ChefScriptException,
    ChefScriptInternalError,
    ChefScriptRuntimeError,
    ChefScriptSyntaxError,
    Position,
    index_to_position,
    pretty_str,
)


class ChefScriptInterpreter:
    recipes: OrderedDict[str, PychefRecipe]
    pos: Position
    filename: str
    code: str

    def __init__(self) -> None:
        self.recipes = OrderedDict()
        self.pos = Position(-1, -1)

    def interpret_file(self, filename: str):
        self.filename = filename
        self.code = Path(filename).resolve().read_text(encoding="utf-8")
        self._interpret(self.code)

    def interpret_stdin(self):
        self.filename = "<stdin>"
        self.code = stdin.read()
        self._interpret(self.code)

    def _interpret(self, code: str):
        if not code.strip():
            return
        elif not code.endswith("\n"):
            code += "\n"
        try:
            try:
                parse_result = ChefScriptParser.parse(code)
            except ParseBaseException as e:
                raise ChefScriptSyntaxError(
                    e.msg,
                    self.filename,
                    Position(e.lineno, e.col),
                )

            for stmt in parse_result:
                if len(stmt) == 1:
                    if isinstance(stmt[0], PychefRecipe):
                        self._add_recipe(stmt[0])
                    elif isinstance(stmt[0], Cook):
                        self._cook(stmt[0])
                    else:
                        raise ChefScriptInternalError("Parser error", self.filename)
                elif len(stmt) >= 2:
                    raise ChefScriptInternalError("Parser error", self.filename)

        except ChefScriptException as e:
            print(e, file=stderr)

        except Exception as e:
            new_e = ChefScriptInternalError(f"Internal error: {e}", self.filename)
            print(new_e, file=stderr)

    def _add_recipe(self, recipe: PychefRecipe):
        self.pos = index_to_position(self.code, recipe.idx)  # type: ignore

        for instruction in recipe.instructions:
            if isinstance(instruction[0], PychefRecipe):
                if instruction[0].name in self.recipes:
                    instruction[0].instructions = self.recipes[
                        instruction[0].name
                    ].instructions
                else:
                    raise ChefScriptRuntimeError(
                        f"Recipe '{instruction[0].name}' "
                        f"used in '{recipe.name}' is not defined yet",
                        self.filename,
                        self.pos,
                    )
        self.recipes[recipe.name] = recipe

    def _cook(self, cook: Cook):
        self.pos = index_to_position(self.code, cook.idx)  # type: ignore

        recipe_name = cook.recipe_name
        if recipe_name not in self.recipes:
            raise ChefScriptRuntimeError(
                f"Recipe '{recipe_name}' is not defined yet", self.filename, self.pos
            )
        recipe: PychefRecipe = self.recipes[recipe_name]

        scale: float = 1

        if cook.scale is not None:
            if not isinstance(cook.scale, float):
                if cook.scale.name not in [i.name for i in recipe.ingredients]:
                    raise ChefScriptRuntimeError(
                        f"Ingredient '{cook.scale.name}' "
                        f"is not in recipe '{recipe_name}'",
                        self.filename,
                        index_to_position(self.code, cook.idx),  # type: ignore
                    )
                for i in recipe.ingredients:
                    if i.name == cook.scale.name:
                        scale = cook.scale.quantity / i.quantity
                        break
            else:
                scale = cook.scale

        print(pretty_str(f"Cooking {recipe_name} with scale {scale:.3f}"))
        print((recipe * scale).pretty_str)
        print("-" * MAX_TERMINTAL_WIDTH)
